from unittest import TestCase

from flask import Blueprint, Flask

from tests.fixtures import clear_dir, init_dir

from ...jwt_auth import USER_AUTH_HEADER_KEY, UserClaims
from ...repositories.project.project import (
    AccessSettings,
    FormStage,
    NotificationTrigger,
    ProjectRepository,
)
from ...repositories.users import LocalUsersRepository
from .role_guard import (
    Guard,
    IdArgSelector,
    PathArgSelector,
    StageIdSelector,
)


class TestRequirementsApi(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_app(self):
        ##Autorization: Bearer <test_email>
        def test_header(email: str):
            return {USER_AUTH_HEADER_KEY: f"Bearer {email}"}

        def test_auth_decoder(email: str):
            email = email.split(" ")[1]
            return UserClaims(claims={"email": email})

        repo = LocalUsersRepository({"tester@abstra.app": ["workflow_viewer"]})
        guard = Guard(repo, test_auth_decoder)

        app = Flask(__name__)
        bp = Blueprint("dummy", __name__)

        @bp.route("/")
        @guard.requires("workflow_viewer")
        def main():
            return {"message": "OK"}

        app.register_blueprint(bp, url_prefix="/")

        client = app.test_client()
        response = client.get("/", headers=test_header("tester@abstra.app"))

        self.assertEqual(response.status, "200 OK")

        client = app.test_client()
        response = client.get("/", headers=test_header("not-a-tester@abstra.app"))

        self.assertEqual(response.status, "403 FORBIDDEN")

    def test_app_protected_by(self):
        ##Autorization: Bearer <test_email>
        def test_header(email: str):
            return {USER_AUTH_HEADER_KEY: f"Bearer {email}"}

        def test_auth_decoder(email: str):
            email = email.split(" ")[1]
            return UserClaims(claims={"email": email})

        JOB_EXECUTOR = "job_executor"
        WORKFLOW_VIEWER = "workflow_viewer"
        ADMIN = "admin"

        TESTER_EMAIL = "tester@abstra.app"
        ADMIN_EMAIL = "admin@abstra.app"
        USER_EMAIL = "user@abstra.app"

        NOT_A_USER_EMAIL = "external@evil.corp"

        repo = LocalUsersRepository(
            {
                TESTER_EMAIL: [WORKFLOW_VIEWER],
                ADMIN_EMAIL: [ADMIN, JOB_EXECUTOR, WORKFLOW_VIEWER],
                USER_EMAIL: [],
            }
        )

        project = ProjectRepository.load()

        public_form = FormStage(
            id="public_form_id",
            path="public_form",
            title="public_form",
            file="public_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=True, required_roles=[]),
        )

        public_form = FormStage(
            id="public_form_id",
            path="public_form",
            title="public_form",
            file="public_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=True, required_roles=[]),
        )

        protected_form = FormStage(
            id="protected_form_id",
            path="protected_form",
            title="protected_form",
            file="protected_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=False, required_roles=[]),
        )

        private_form = FormStage(
            id="private_form_id",
            path="private_form",
            title="private_form",
            file="private_form.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
            access_control=AccessSettings(is_public=False, required_roles=[ADMIN]),
        )

        project.forms.append(public_form)
        project.forms.append(protected_form)
        project.forms.append(private_form)

        project.kanban.access_control = AccessSettings(
            is_public=False, required_roles=[WORKFLOW_VIEWER]
        )

        ProjectRepository.save(project)

        guard = Guard(repo, test_auth_decoder)

        app = Flask(__name__)
        bp = Blueprint("dummy", __name__)

        @bp.route("/kanban")
        @guard.by(StageIdSelector("kanban"))
        def kanban():
            return {"message": "OK"}

        @bp.route("/protected_form_id")
        @guard.by(StageIdSelector("protected_form_id"))
        def protected_form_route():
            return {"message": "OK"}

        @bp.route("/_pages/<string:path>")
        @guard.by(PathArgSelector("path"))
        def path_stage(path: str):
            return {"message": "OK", "path": path}

        @bp.route("/stage/<path:id>/start")
        @guard.by(IdArgSelector("id"))
        def id_stage(id: str):
            return {"message": "OK", "id": id}

        app.register_blueprint(bp, url_prefix="/")

        ## Logged in users
        test_matrix = [
            {
                "email": TESTER_EMAIL,
                "test_routes": {
                    "/kanban": "200 OK",
                    "/protected_form_id": "200 OK",
                    "/_pages/public_form": "200 OK",
                    "/_pages/protected_form": "200 OK",
                    "/_pages/private_form": "403 FORBIDDEN",
                    "/stage/public_form_id/start": "200 OK",
                    "/stage/protected_form_id/start": "200 OK",
                    "/stage/private_form_id/start": "403 FORBIDDEN",
                },
            },
            {
                "email": ADMIN_EMAIL,
                "test_routes": {
                    "/kanban": "200 OK",
                    "/protected_form_id": "200 OK",
                    "/_pages/public_form": "200 OK",
                    "/_pages/protected_form": "200 OK",
                    "/_pages/private_form": "200 OK",
                    "/stage/public_form_id/start": "200 OK",
                    "/stage/protected_form_id/start": "200 OK",
                    "/stage/private_form_id/start": "200 OK",
                },
            },
            {
                "email": USER_EMAIL,
                "test_routes": {
                    "/kanban": "403 FORBIDDEN",
                    "/protected_form_id": "200 OK",
                    "/_pages/public_form": "200 OK",
                    "/_pages/protected_form": "200 OK",
                    "/_pages/private_form": "403 FORBIDDEN",
                    "/stage/public_form_id/start": "200 OK",
                    "/stage/protected_form_id/start": "200 OK",
                    "/stage/private_form_id/start": "403 FORBIDDEN",
                },
            },
            {
                "email": NOT_A_USER_EMAIL,
                "test_routes": {
                    "/kanban": "403 FORBIDDEN",
                    "/protected_form_id": "403 FORBIDDEN",
                    "/_pages/public_form": "200 OK",
                    "/_pages/protected_form": "403 FORBIDDEN",
                    "/_pages/private_form": "403 FORBIDDEN",
                    "/stage/public_form_id/start": "200 OK",
                    "/stage/protected_form_id/start": "403 FORBIDDEN",
                    "/stage/private_form_id/start": "403 FORBIDDEN",
                },
            },
        ]

        client = app.test_client()
        for case in test_matrix:
            for route, status in case["test_routes"].items():
                response = client.get(route, headers=test_header(case["email"]))
                self.assertEqual(
                    response.status, status, f"Failed for {route} with {case['email']}"
                )

        ## Logged off user
        test_matrix = {
            "/kanban": "401 UNAUTHORIZED",
            "/protected_form_id": "401 UNAUTHORIZED",
            "/_pages/public_form": "200 OK",
            "/_pages/protected_form": "401 UNAUTHORIZED",
            "/_pages/private_form": "401 UNAUTHORIZED",
            "/stage/public_form_id/start": "200 OK",
            "/stage/protected_form_id/start": "401 UNAUTHORIZED",
            "/stage/private_form_id/start": "401 UNAUTHORIZED",
        }

        for route, status in test_matrix.items():
            response = client.get(route, headers={})
            self.assertEqual(response.status, status, f"Failed for {route}")



import flask

from ...usage import usage
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_ai", __name__)

    @bp.post("/message")
    @usage
    def _get_next_message():
        body = flask.request.json
        if not body:
            flask.abort(400)

        streamer = controller.send_ai_message(
            body["messages"], body["runtime"], body["threadId"]
        )

        if streamer is None:
            flask.abort(403)

        return flask.Response(streamer, mimetype="text/event-stream")

    @bp.post("/thread")
    @usage
    def _create_thread():
        thread = controller.create_thread()
        if not thread:
            flask.abort(403)
        return thread

    return bp



import datetime, re
from ..widget_base import Input
from typing import Optional


class TimeInput(Input):
    type = 'time-input'
    empty_value = None

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.value = props.get('initial_value', self.empty_value)
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.format = props.get('format', '24hs')
        self.disabled = props.get('disabled', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'format': self.format, 'hint': self.hint, 'value': self.
            serialize_value(), 'required': self.required, 'fullWidth': self
            .full_width, 'disabled': self.disabled, 'errors': self.errors}

    def serialize_value(self) ->Optional[dict]:
        if isinstance(self.value, str) and re.match('^\\d{2}:\\d{2}$', self
            .value):
            return {'hour': int(self.value.split(':')[0]), 'minute': int(
                self.value.split(':')[1])}
        if isinstance(self.value, datetime.time):
            return {'hour': self.value.hour, 'minute': self.value.minute}
        return {'hour': 0, 'minute': 0}

    def parse_value(self, value) ->Optional[datetime.time]:
        if value is None or value['hour'] is None or value['minute'] is None:
            return self.empty_value
        return datetime.time(value['hour'], value['minute']
            ) if value else self.empty_value



from ..widget_base import Output
from urllib.parse import quote
from validators import url


class IframeOutput(Output):
    type = 'iframe-output'

    def __init__(self, url_or_html: str, **kwargs):
        self.set_props(dict(url_or_html=url_or_html, **kwargs))

    def set_props(self, props):
        url_or_html = props.get('url_or_html', '')
        if url(url_or_html):
            self.url = url_or_html
        else:
            self.url = f'data:text/html,{quote(url_or_html)}'
        self.width = props.get('width', '800')
        self.height = props.get('height', '600')
        self.full_width = props.get('full_width', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'url': self.url, 'width': self.width,
            'height': self.height, 'fullWidth': self.full_width}



import sys

import flask

from ...debugger.vscode import (
    configure_launch_json,
    configure_settings_json,
    is_client_connected,
    is_launch_json_configured,
    is_settings_json_set,
    start_debugger,
)
from ...settings import Settings
from ...usage import usage
from .main import MainController


# Debugger
def get_debugger_status():
    return dict(
        port=start_debugger(),
        connected=is_client_connected(),
        is_launch_json_configured=is_launch_json_configured()
        and is_settings_json_set(),
        interpreter=sys.executable,
        root=str(Settings.root_path),
    )


def create_vscode_launch():
    configure_launch_json()
    configure_settings_json()
    return {"status": "ok"}


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_debugger", __name__)

    @bp.get("/")
    def _get_debugger_status():
        return get_debugger_status()

    @bp.post("/vscode-launch")
    @usage
    def _create_vscode_launch():
        return create_vscode_launch()

    return bp



from unittest import TestCase

from tests.fixtures import clear_dir, init_dir

from ...repositories.project.project import (
    FormStage,
    JobStage,
    NotificationTrigger,
    ProjectRepository,
)
from ...repositories.stage_run import LocalStageRunRepository
from .kanban import DataRequest, KanbanController, KanbanData, StageRunCard


class KanbanTests(TestCase):
    def setUp(self):
        self.root = init_dir()

        self.stage_run_repository = LocalStageRunRepository()
        self.project_repository = ProjectRepository
        self.controller = KanbanController(
            stage_run_repository=self.stage_run_repository,
            project_repository=self.project_repository,
        )

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_get_data_empty(self):
        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data, KanbanData(stage_run_cards=[], not_found_stages=[], total_count=0)
        )

    def test_get_data_stage_no_filters(self):
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        project.add_stage(job)
        project.add_stage(form)
        ProjectRepository.save(project)
        job_stage_run = self.stage_run_repository.create_initial("job")
        form_stage_run = self.stage_run_repository.create_initial("form")

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                stage_run_cards=[
                    StageRunCard(
                        form_stage_run.id,
                        created_at=form_stage_run.created_at,
                        updated_at=form_stage_run.updated_at,
                        status=form_stage_run.status,
                        content=[],
                        stage=form_stage_run.stage,
                    ),
                    StageRunCard(
                        job_stage_run.id,
                        created_at=job_stage_run.created_at,
                        updated_at=job_stage_run.updated_at,
                        status=job_stage_run.status,
                        content=[],
                        stage=job_stage_run.stage,
                    ),
                ],
                not_found_stages=[],
                total_count=2,
            ),
        )

    def test_get_data_with_stage_filter(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        project.add_stage(job)
        project.add_stage(form)
        ProjectRepository.save(project)
        job_stage_run = self.stage_run_repository.create_initial("job")
        self.stage_run_repository.create_initial("form")

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "limit": 10,
                    "offset": 0,
                    "filter": {
                        "stage": ["job"],
                    },
                }
            )
        )

        self.assertEqual(
            data,
            KanbanData(
                stage_run_cards=[
                    StageRunCard(
                        job_stage_run.id,
                        created_at=job_stage_run.created_at,
                        updated_at=job_stage_run.updated_at,
                        status=job_stage_run.status,
                        content=[],
                        stage=job_stage_run.stage,
                    ),
                ],
                not_found_stages=[],
                total_count=1,
            ),
        )

    def test_get_job(self):
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(job)
        ProjectRepository.save(project)

        rtvd = self.controller.get_job("job")

        self.assertEqual(rtvd, job)



import json

import flask
import flask_sock

from ...execution import FormExecution
from ...execution.execution import RequestData
from ...execution.stage_run_manager import (
    AttachedStageRunManager,
    DetachedStageRunManager,
)
from ...logger import AbstraLogger
from ...repositories import stage_run_repository_factory
from ...usage import usage
from ...utils import is_it_true
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_forms", __name__)
    sock = flask_sock.Sock(bp)

    @sock.route("/socket")
    def _websocket(conn: flask_sock.Server):
        request_data = RequestData(
            query_params=flask.request.args,
            body=flask.request.get_data(as_text=True),
            headers=flask.request.headers,
            method=flask.request.method,
        )

        try:
            id = flask.request.args.get("id")
            if id is None:
                return conn.close(reason=400, message="No path")

            form = controller.get_form(id)
            if not form:
                return conn.close(reason=404, message="Not found")

            if flask.request.args.get("detached", default=False, type=is_it_true):
                stage_run_repository = stage_run_repository_factory()
                data = json.loads(controller.read_test_data())
                stage_run_manager = DetachedStageRunManager(
                    stage_run_repository, data=data
                )
            else:
                stage_run_manager = AttachedStageRunManager(
                    controller.stage_run_repository
                )

            execution = FormExecution(
                is_initial=controller.is_initial(form.id),
                request=request_data,
                connection=conn,
                stage=form,
                execution_logs_repository=controller.execution_logs_repository,
                execution_repository=controller.execution_repository,
                stage_run_manager=stage_run_manager,
            )

            execution.run()
        except Exception as e:
            AbstraLogger.capture_exception(e)
        finally:
            conn.close(message="Done")

    @bp.get("/")
    @usage
    def _get_forms():
        return [f.editor_dto for f in controller.get_forms()]

    @bp.get("/<path:id>")
    @usage
    def _get_form(id: str):
        form = controller.get_form(id)
        if not form:
            flask.abort(404)
        return form.editor_dto

    @bp.delete("/<path:id>")
    @usage
    def _delete_form(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_form(id, remove_file)
        return {"success": True}

    @bp.post("/")
    @usage
    def _create_form():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        form = controller.create_form(title, file, workflow_position)
        return form.editor_dto

    @bp.put("/<path:id>")
    @usage
    def _update_form(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        form = controller.update_stage(id, data)
        return form.editor_dto if form else None

    return bp



from typing import List

from ..linter import LinterRule
from .duplicate_package_in_requirements import DuplicatePackagesInRequirements
from .env_file_changed import EnvFileChanged
from .env_in_bundle import EnvInBundle
from .missing_abstra_in_requirements import MissingAbstraInRequirements
from .missing_entrypoint import MissingEntrypoint
from .missing_env import MissingEnv
from .missing_packages_in_requirements import MissingPackagesInRequirements
from .new_version_of_abstra_available import NewVersionOfAbstraAvailable
from .syntax_errors import SyntaxErrors
from .unset_get_data import UnsetGetData
from .unset_match_thread import UnsetMatchThread

rules: List[LinterRule] = [
    EnvInBundle(),
    MissingPackagesInRequirements(),
    SyntaxErrors(),
    MissingEntrypoint(),
    MissingAbstraInRequirements(),
    DuplicatePackagesInRequirements(),
    MissingEnv(),
    NewVersionOfAbstraAvailable(),
    UnsetGetData(),
    UnsetMatchThread(),
    EnvFileChanged(),
]



from ....repositories import connectors_api_http_client


class ConnectorExecutionError(Exception):  # public api
    def __init__(self, name: str, method_name: str, errors: list) -> None:
        self.name = name
        self.method_name = method_name
        self.errors = errors


class Method:
    def __init__(self, name: str, connector: "Connector") -> None:
        self.name = name
        self.connector = connector

    def __call__(self, **kwargs):  # public api
        r = connectors_api_http_client.execute(
            name=self.connector.name, method=self.name, params=kwargs
        )

        if not r.ok:
            raise Exception(f"Error executing method {self.name}: {r.text}")

        result = r.json()
        if result["errors"]:
            raise ConnectorExecutionError(
                self.connector.name, self.name, result["errors"]
            )

        return result["returns"]


class Connector:
    def __init__(self, name: str) -> None:
        self.name = name

    def get_method(self, method_name: str) -> "Method":  # public api
        return Method(method_name, self)

    def execute(self, method_name: str, **kwargs):  # public api
        return self.get_method(method_name)(**kwargs)

    def __getattr__(self, method_name: str) -> "Method":
        return self.get_method(method_name)


def get_connector(name: str) -> "Connector":  # public api
    return Connector(name)



import flask

from ...repositories import authn_repository
from ...utils.email import is_valid_email


def get_player_bp():
    bp = flask.Blueprint("auth", __name__)

    @bp.post("/authenticate")
    def _authenticate():
        data = flask.request.get_json(force=True)
        if not data:
            return flask.abort(400)

        email = data.get("email")
        if not email or not is_valid_email(email):
            return flask.abort(400)

        ok = authn_repository.authenticate(email)
        if not ok:
            return flask.abort(400)

        return {"ok": True}

    @bp.post("/verify")
    def _verify():
        data = flask.request.get_json(force=True)
        if not data:
            return flask.abort(400)

        email = data.get("email")
        token = data.get("token")
        if not email or not token:
            return flask.abort(400)

        jwt = authn_repository.verify(email, token)
        if not jwt:
            return flask.abort(404)

        return {"jwt": jwt}

    @bp.post("/oidc-verify")
    def _oidc_access():
        data = flask.request.get_json(force=True)
        if not data:
            return flask.abort(400)

        access_token = data.get("access_token")
        if not access_token:
            return flask.abort(400)

        jwt = authn_repository.oidc_access(access_token)
        if not jwt:
            return flask.abort(404)

        return {"jwt": jwt}

    return bp



import flask

from ...repositories import users_repository
from ...usage import usage
from ..guards.role_guard import Guard
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("access_control", __name__)

    @bp.get("/")
    @usage
    def _get_project_level_access_control():
        return controller.list_access_controls()

    @bp.put("/")
    @usage
    def _update_access_controls():
        changes = flask.request.json
        if not changes:
            flask.abort(400)

        access_control = controller.update_access_controls(changes)
        return access_control if access_control else None

    return bp


def get_player_bp():
    guard = Guard(users_repository)

    bp = flask.Blueprint("player_access_control", __name__)

    @bp.get("/allow/<string:path>")
    def _get_allow_status_by_path(path: str):
        authHeader = flask.request.headers.get("Authorization")
        return guard.allow(path, authHeader).to_dict()

    return bp



from typing import Dict, List, Optional, TypedDict

import flask

from ...logger import AbstraLogger
from ...repositories.project.project import (
    ConditionStage,
    FormStage,
    HookStage,
    IteratorStage,
    JobStage,
    NotificationTrigger,
    Project,
    ProjectRepository,
    ScriptStage,
    WorkflowStage,
    WorkflowTransition,
)
from ...usage import usage
from .main import MainController, UnknownNodeTypeError


# workflow visual editor
def get_workflow():
    project = ProjectRepository.load()
    return _make_workflow_dto(project)


def build_adjacency_list() -> Dict[str, List[str]]:
    workflow = get_workflow()
    transitions = workflow["transitions"]
    adj = {}
    for t in transitions:
        if t["sourceStageId"] not in adj:
            adj[t["sourceStageId"]] = []
        adj[t["sourceStageId"]].append(t["targetStageId"])
    return adj


def get_initial_stages():
    workflow = get_workflow()
    stages = workflow["stages"]
    target_stage_ids = set([t["targetStageId"] for t in workflow["transitions"]])
    return [s for s in stages if s["id"] not in target_stage_ids]


def get_path(n: int):
    adj = build_adjacency_list()
    initial_stages = get_initial_stages()
    longest_path = []

    def dfs(stage_id: str, current_path: List[str]):
        if stage_id in current_path:
            return
        current_path = current_path.copy()
        current_path.append(stage_id)
        if len(current_path) > len(longest_path):
            longest_path.clear()
            longest_path.extend(current_path)
        for next_stage_id in adj.get(stage_id, []):
            dfs(next_stage_id, current_path)

    for stage in initial_stages:
        dfs(stage["id"], [])

    return longest_path[:n]


class StageDTO(TypedDict):
    id: str
    type: str
    title: str
    position: Dict[str, float]
    props: Dict[str, Optional[str]]


def make_stage_dto(stage: WorkflowStage) -> StageDTO:
    filename = None
    if isinstance(stage, (HookStage, ScriptStage, FormStage, JobStage)):
        filename = stage.file
    variable_name = None
    if isinstance(stage, (IteratorStage, ConditionStage)):
        variable_name = stage.variable_name
    path = None
    if isinstance(stage, (FormStage, HookStage)):
        path = stage.path
    return StageDTO(
        id=stage.id,
        type=stage.type_name + "s",
        title=stage.title,
        position=dict(
            x=stage.workflow_position[0],
            y=stage.workflow_position[1],
        ),
        props={
            "filename": filename,
            "variableName": variable_name,
            "path": path,
        },
    )


def _make_workflow_dto(project: Project):
    stages = []
    for stage in project.workflow_stages:
        stage_dto = make_stage_dto(stage)

        if isinstance(stage, IteratorStage) or isinstance(stage, ConditionStage):
            stage_dto["props"]["variableName"] = stage.variable_name

        if (
            isinstance(stage, FormStage)
            or isinstance(stage, ScriptStage)
            or isinstance(stage, HookStage)
            or isinstance(stage, JobStage)
        ):
            stage_dto["props"]["filename"] = stage.file

        if isinstance(stage, FormStage) or isinstance(stage, HookStage):
            stage_dto["props"]["path"] = stage.path

        stages.append(stage_dto)

    transitions = []
    for stage in project.workflow_stages:
        for transition in stage.workflow_transitions:
            props: dict = {"conditionValue": None}

            if transition.condition_value is not None:
                props["conditionValue"] = transition.condition_value

            transitions.append(
                dict(
                    id=transition.id,
                    sourceStageId=stage.id,
                    targetStageId=transition.target_id,
                    type=transition.type,
                    props=props,
                )
            )

    return dict(
        stages=stages,
        transitions=transitions,
    )


def update_workflow(workflow_state_dto: dict):
    project = ProjectRepository.load()

    for stage_dto in workflow_state_dto["stages"]:
        stage = project.get_stage(stage_dto["id"])
        if stage is not None:
            stage.workflow_position = (
                stage_dto["position"]["x"],
                stage_dto["position"]["y"],
            )
            if not isinstance(stage, ConditionStage) and not isinstance(
                stage, IteratorStage
            ):
                stage.title = stage_dto["title"]
            if isinstance(stage, IteratorStage) or isinstance(stage, ConditionStage):
                stage.variable_name = stage_dto["props"]["variableName"]
            if (
                isinstance(stage, FormStage)
                or isinstance(stage, ScriptStage)
                or isinstance(stage, HookStage)
                or isinstance(stage, JobStage)
            ):
                stage.file = f"{stage_dto['props']['filename']}"
            if isinstance(stage, FormStage) or isinstance(stage, HookStage):
                stage.path = stage_dto["props"]["path"]
        elif stage_dto["type"] == "forms":
            stage = FormStage(
                id=stage_dto["id"],
                path=stage_dto["props"]["path"],
                file=f"{stage_dto['props']['filename']}",
                title=stage_dto["title"],
                workflow_transitions=[],
                is_initial=False,
                workflow_position=(
                    stage_dto["position"]["x"],
                    stage_dto["position"]["y"],
                ),
                notification_trigger=NotificationTrigger(
                    variable_name="assignee_emails",
                    enabled=False,
                ),
            )
            project.forms.append(stage)
        elif stage_dto["type"] == "scripts":
            stage = ScriptStage(
                file=f"{stage_dto['props']['filename']}",
                is_initial=False,
                id=stage_dto["id"],
                title=stage_dto["title"],
                workflow_position=(
                    stage_dto["position"]["x"],
                    stage_dto["position"]["y"],
                ),
                workflow_transitions=[],
            )
            project.scripts.append(stage)
        elif stage_dto["type"] == "hooks":
            stage = HookStage(
                id=stage_dto["id"],
                enabled=True,
                file=f"{stage_dto['props']['filename']}",
                is_initial=False,
                path=stage_dto["id"],
                title=stage_dto["title"],
                workflow_position=(
                    stage_dto["position"]["x"],
                    stage_dto["position"]["y"],
                ),
                workflow_transitions=[],
            )
            project.hooks.append(stage)
        elif stage_dto["type"] == "jobs":
            stage = JobStage(
                file=f"{stage_dto['props']['filename']}",
                id=stage_dto["id"],
                schedule="0 * * * *",
                title=stage_dto["title"],
                workflow_position=(
                    stage_dto["position"]["x"],
                    stage_dto["position"]["y"],
                ),
                workflow_transitions=[],
            )
            project.jobs.append(stage)
        elif stage_dto["type"] == "iterators":
            props = stage_dto.get("props", {})
            variable_name = props.get("variableName", "")

            stage = IteratorStage(
                id=stage_dto["id"],
                workflow_position=(
                    stage_dto["position"]["x"],
                    stage_dto["position"]["y"],
                ),
                variable_name=variable_name,
                workflow_transitions=[],
            )
            project.add_stage(stage)
        elif stage_dto["type"] == "conditions":
            props = stage_dto.get("props", {})
            variable_name = props.get("variableName", "")

            stage = ConditionStage(
                id=stage_dto["id"],
                workflow_position=(
                    stage_dto["position"]["x"],
                    stage_dto["position"]["y"],
                ),
                variable_name=variable_name,
                workflow_transitions=[],
            )
            project.add_stage(stage)
        else:
            raise UnknownNodeTypeError(stage_dto["type"])

    # Delete removed stages
    for stage in project.workflow_stages:
        if stage.id not in [s["id"] for s in workflow_state_dto["stages"]]:
            project.delete_stage(stage.id)

    for stage in project.workflow_stages:
        stage.workflow_transitions = []
        for transition in workflow_state_dto["transitions"]:
            props = transition.get("props", {})
            condition_value = props.get("conditionValue", None)
            if transition["sourceStageId"] == stage.id:
                target = project.get_stage(transition["targetStageId"])
                assert target is not None
                stage.workflow_transitions.append(
                    WorkflowTransition(
                        id=transition["id"],
                        type=transition["type"],
                        target_id=transition["targetStageId"],
                        target_type=target.type_name + "s",
                        condition_value=condition_value,
                    )
                )

    ProjectRepository.save(project)

    return _make_workflow_dto(project)


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_workflows", __name__)

    # 1s pooling in this route
    @bp.get("/")
    def _load_workflow():
        try:
            return get_workflow()
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return str(e), 500

    @bp.put("/")
    @usage
    def _update_workflow():
        try:
            payload = flask.request.json
            if payload is None:
                raise Exception("No payload found")
            new_state = update_workflow(payload)
            return new_state
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return str(e), 500

    return bp



from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional, TypedDict, TypeVar, Union


class Widget(ABC):
    @abstractmethod
    def render(self, ctx: dict):
        raise NotImplementedError("render not implemented")

    @abstractmethod
    def set_props(self, props):
        raise NotImplementedError("set_props not implemented")


class Output(Widget):
    type: str

    def has_errors(self):
        return False


class Input(Widget):
    type: str
    errors: List[str] = (lambda: [])()
    empty_value: Any = None
    required: Optional[Union[bool, str]] = None

    def __init__(self, key: str) -> None:
        super().__init__()
        self.key = key

    def set_value(self, value, set_errors=False):
        old_value = self.value if hasattr(self, "value") else None
        self.value = value

        if set_errors and old_value != value:
            self.set_errors()

    def validate(self) -> List[str]:
        errors = []
        for validator in self.validators:
            errors += validator()
        return errors

    def set_errors(self):
        self.errors = self.validate()

    def has_errors(self):
        return len(self.errors) > 0

    def is_value_unset(self) -> bool:
        return (
            self.value is None
            or self.value == ""
            or self.value == []
            or (isinstance(self.empty_value, bool) and self.value is False)
            or self.value == {}
        )

    def _validate_required(self) -> List[str]:
        if self.required and self.is_value_unset():
            if isinstance(self.required, str):
                return [self.required]

            return ["i18n_error_required_field"]

        return []

    @property
    def validators(self) -> List[Callable[[], List[str]]]:
        return [self._validate_required]

    def parse_value(self, value):
        _value = value
        return _value if _value is not None else self.empty_value

    @abstractmethod
    def serialize_value(self):
        raise NotImplementedError("serialize_value not implemented")


class LabelValueOption(TypedDict):
    label: str
    value: object


AbstraOption = Union[str, LabelValueOption]


class OptionsHandler:
    def __init__(self, options: Union[List[str], List[LabelValueOption]]) -> None:
        self.options = options

        self._mappedOptions = {
            str(index): {
                "label": option if isinstance(option, str) else option["label"],
                "value": option if isinstance(option, str) else option["value"],
            }
            for index, option in enumerate(options)
        }

    def ids_from_values(self, values: List[object]) -> List[str]:
        value = [self._id_from_value(value) for value in values]
        return [v for v in value if v is not None]

    def values_from_ids(self, ids: Union[List[str], None]) -> List[object]:
        if ids is None:
            return []

        return [self._value_from_id(id) for id in ids]

    def _value_from_id(self, id: str) -> object:
        return self._mappedOptions[id]["value"]

    def _id_from_value(self, value: object) -> Optional[str]:
        for mapKey, mapValue in self._mappedOptions.items():
            if mapValue["value"] == value:
                return mapKey
        return None

    def serialized_options(self) -> List[LabelValueOption]:
        return [
            {"label": v["label"], "value": k} for k, v in self._mappedOptions.items()
        ]


T = TypeVar("T")


class MultipleHandler:
    def __init__(
        self,
        multiple: bool,
        min: Optional[int] = None,
        max: Optional[int] = None,
        required: Optional[Union[bool, str]] = None,
    ) -> None:
        self.multiple = multiple
        self.min = min
        self.max = max
        self.required = required

    def _is_value_unset(self, value: Union[object, List[object]]) -> bool:
        return value is None or value == []

    def validate(self, value: Union[object, List[object]]) -> List[str]:
        if self.required and self._is_value_unset(value):
            if isinstance(self.required, str):
                return [self.required]

            return ["i18n_error_required_field"]

        if not self.multiple or not isinstance(value, list):
            return []

        if self.min is not None and len(value) < self.min:
            return ["i18n_error_min_list"]

        if self.max is not None and len(value) > self.max:
            return ["i18n_error_max_list"]

        return []

    def value_to_list_or_value(self, value: List[T]) -> Union[List[T], T, None]:
        if not isinstance(value, list) and value is not None:
            raise Exception("Non-list value received")

        if self.multiple:
            return value

        return value[0] if len(value) else None

    def value_to_list(self, value: Union[List[T], T]) -> List[T]:
        if isinstance(value, list):
            return value
        if value is None:
            return []
        return [value]


class NumberValueHandler:
    def __init__(
        self,
        min: Optional[Union[int, float]] = None,
        max: Optional[Union[int, float]] = None,
        required: Optional[Union[bool, str]] = None,
    ) -> None:
        self.min = min
        self.max = max
        self.required = required

    def validate(self, value: Union[int, float]) -> List[str]:
        errors = []
        if self.required and value is None:
            if isinstance(self.required, str):
                errors.append(self.required)
            else:
                errors.append("i18n_error_required_field")

        if self.min is not None and value < self.min:
            errors.append("i18n_error_min_number")

        if self.max is not None and value > self.max:
            errors.append("i18n_error_max_number")

        return errors



import abc
import typing

import requests

from ..credentials import resolve_headers
from ..environment import CLOUD_API_CLI_URL, SIDECAR_HEADERS, SIDECAR_URL


class ConnectorsApiHttpClient(abc.ABC):
    execute_url: str

    def __init__(self, base_url: str) -> None:
        self.execute_url = f"{base_url}/connectors/execute"

    def execute(self, name: str, method: str, params: typing.Dict) -> requests.Response:
        raise NotImplementedError()


class ProductionConnectorsApiHttpClient(ConnectorsApiHttpClient):
    def execute(self, name: str, method: str, params: typing.Dict) -> requests.Response:
        body = {"name": name, "method": method, "params": params}
        return requests.post(self.execute_url, headers=SIDECAR_HEADERS, json=body)


class LocalConnectorsApiHttpClient(ConnectorsApiHttpClient):
    def execute(self, name: str, method: str, params: typing.Dict) -> requests.Response:
        body = {"name": name, "method": method, "params": params}
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to execute a connector method")
        return requests.post(self.execute_url, headers=headers, json=body)


def connectors_api_http_client_factory() -> ConnectorsApiHttpClient:
    if SIDECAR_URL is None:
        return LocalConnectorsApiHttpClient(CLOUD_API_CLI_URL)
    else:
        return ProductionConnectorsApiHttpClient(SIDECAR_URL)



import flask

from ...repositories import roles_repository
from ...usage import usage


def get_editor_bp():
    bp = flask.Blueprint("roles", __name__)

    @bp.get("/")
    @usage
    def _get_project_roles():
        return roles_repository.get_roles()

    return bp
